local class = require 'lib.middleclass'


Boss1 = class('Boss1')

function Boss1:initialize(_x, _y, _size) 

	self.name = "enemy"
	
	self.id = math.random(1,100000) -- for testing purposes
		
	if (_size == nil) then
		self.h = 100
		self.w = self.h
	else
		self.h = _size
		self.w = _size
	end 
	
	self.x = _x
	self.y = _y
	self.angle = 0
	self.firingspeed = 0
	self.rotatespeed = 2.5
	self.accelerationX = 0
	self.accelerationY = 0
	self.friction = 0.99
	self.thrust = math.random(300,400)
	self.maxspeed = math.random(500,800)
	self.speedX = 0 --math.random(-350, 350)
	self.speedY = 0 --math.random(-350, 350)
	
	local modifier = game:getStageModifier() 
	
	-- this enemy doesn't split atm.
	self.maxhealth = self.h * modifier * 10
	self.health =  self.maxhealth 
	
	self.pointvalue = 5000 * modifier --point value
	
	-- init weapon
	local pmax = WeaponModifiers:getBombPropMax() 
	-- note, DAMAGE MODIFIER set to 0 to keep in line with other enemies max bomb dmg (100)
	local weapon_modifiers = WeaponModifiers(pmax.fs, pmax.ls, 0, pmax.ft, pmax.b) -- firingspeed, lifespan, damage, firingtimer, bounces || try unpack(pmax) ??
	
	self.weaponBomb = Weapon("bomb", self.name, weapon_modifiers)
	self.weaponBomb.armed = true
	self.weaponSpreadBomb = Weapon("spreadbomb", self.name, weapon_modifiers)
	self.weaponSpreadBomb.armed = true
	
	self.weapon = self.weaponBomb
	
	
	self.min_fire_range = 0
	self.max_fire_range = 1200
	
	self.spawntimerreset = 10
	self.spawntimermax = self.spawntimerreset
	self.spawntimer = 0
	
	self.angrytimermax = math.random(5,10)
	self.angrytimer = 0
	
	
	self.state = "idle" -- idle, angry, fleeing
	self.path = nil
	self.needsGoal = true
	self.goal = nil
	
end



function Boss1:updateSpawningTimer(sh, dt) -- enemy1 spawns additional enemies intermittently (atm just enemy2)
	local sx, sy = sh:center()

	if self.spawntimer >= self.spawntimermax then
		
		-- spawn an enemy2 entity outside of boss1's collision area, at his current facing/angle.
		local a = self.angle
		local cosa = math.cos(a)
		local sina =  math.sin(a)
		local midx, midy = sx + self.h + 30, sy -- add 60 because that's the largest radius an enemy2 can spawn with. I dont want a enemy/enemy collision when enemy2 spawns.
		local x = sx + (midx - sx) * cosa - (midy - sy) * sina
		local y = sy + (midx - sx) * sina + (midy - sy) * cosa
		
		-- increase spawning frequency to prevent farming 
		if self.spawntimermax > 1 then 
			self.spawntimermax = self.spawntimermax - 1
			local newEntity = Entities("enemy2", x, y)
			newEntity:spawn()
		else
			self.spawntimermax = self.spawntimerreset -- reset spawn timer to default
			-- BUT SPAWN ANOTHER BOSS1 MUHAHA
			local newEntity = Entities("boss1", x, y)
			newEntity:spawn()
		end
		
		self.spawntimer = 0
	else 
		self.spawntimer = self.spawntimer + dt
	end
end

function Boss1:updateAngryTimer(sh, dt) -- enemy1 spawns additional enemies intermittently (atm just enemy2)
	local sx, sy = sh:center()

	if self.angrytimer >= self.angrytimermax then
		-- spawn an enemy2 entity
		self.state = "idle"
		self.angrytimer = 0
	else 
		self.angrytimer = self.angrytimer + dt
	end

end


function Boss1:draw(sh)
		
	self:setColor()	
	
	local sx, sy = sh:center()
	local r,g,b = love.graphics.getColor()
	
	love.graphics.setColor(r,g,b,15) 

	-- draw range radius
	local thickness = self.max_fire_range - self.min_fire_range
	love.graphics.setLineWidth( thickness )
	love.graphics.circle('line', sx, sy, self.min_fire_range + (thickness/2))
	love.graphics.setLineWidth(1)
	
	-- draw shape (self)
	love.graphics.setColor(r,g,b,255)
	sh:draw('fill')
		
	-- DRAW HEALTH BAR
	local percent = (self.health / self.maxhealth) * 100
	local lx, ly = sx - self.h, sy - self.h - 20
	local lx2, ly2 = sx + self.h, ly
	local dist = math.dist(lx, ly, lx2, ly2)
	local line_dist = dist * (percent/100)
	love.graphics.line(lx, ly, lx + line_dist, ly2)
	
	-- PRINT SELECTED WEAPON
	love.graphics.setColor(255,255,255)
	local font_w = mainFont:getWidth(self.weapon.weapon_type) -- not sure if this detects scale too. seems to return half as wide as i sould expect.
	love.graphics.print(self.weapon.weapon_type, sx - font_w , ly - 50, 0, 2, 2)
	

	if (game.settings.debug) then
		-- DEBUG TESTING
		local linex, liney
		linex = sx + 200 * math.cos(self.angle)
		liney = sy + 200 * math.sin(self.angle)
		love.graphics.line(sx, sy, linex, liney)  -- draw line to indicate enemy's current angle.
		
		-- DEBUG TESTING
		-- draw path tiles
		if self.path ~= nil then
			for node, count in self.path:nodes() do
				game:drawTileBorderRowCol(node.y, node.x)		  
				local nodex, nodey = world:getTileXY(node.y, node.x)
				if count == 2 then
					love.graphics.setColor(255, 207, 87)
				end
				love.graphics.circle('fill', nodex, nodey, 10) -- draw dots on tiles
				
			end
		end
		
	end
	
end

function Boss1:update(sh, dt)
	local px, py = player1:center()
	local sx, sy = sh:center()	


	if self.state == "angry" then
		self:updateAttackingState(sh,dt) -- does it's own updateAngle() call
		self:updateAngryTimer(sh, dt)
	elseif self.state == "idle" then
		
		self:updateAngle(sx, sy, px, py, dt) -- keep facing player.
		self:updateMovement(sh, dt) -- still update movement so he doesnt just freeze in place.
		self:updateSpawningTimer(sh, dt) -- only update spawning timer when idle.
		
	elseif self.state == "fleeing" then
		
--		self:updateAngle(sx, sy, px, py, dt) -- TEMP. he wont want to face player while fleeing. 
		self:updateFleeingState(sh, dt)
	end	
	
	
	
end
  
  function Boss1:updateAngle(sx, sy, px, py, dt)
  
		local newangle = math.angle(sx, sy, px, py)
		
		-- determine shortest distance required to turn to newangle
		-- use degrees 
		local rotation_direction 
		if (math.deg(self.angle)-math.deg(newangle)+360) % 360 > 180 then 
			rotation_direction = "positive"
		else
			rotation_direction = "negative"
		end
		-- rotate in the appropriate direction			
		if rotation_direction == "positive" then
			self.angle = self.angle + self.rotatespeed *  dt
		elseif rotation_direction == "negative" then
			self.angle = self.angle - self.rotatespeed * dt
		end
		
  end
  
  
  function Boss1:updateFleeingState(sh, dt)
	local px, py = player1:center()
	local sx, sy = sh:center()			
	
	
	
	
		-- PATHFIND!
		local s_row, s_col = world:getTilePosition(sx, sy, true) -- get self tile position in grid
		local p_row, p_col = world:getTilePosition(px, py, true) -- get player tile position in grid.
		local destination = {row = 0, col = 0, distance = 0}

		--if self.needsGoal then
		
			local corners = {}
			corners.ul = { row = 0, col = 0 } -- upper left corner
			corners.bl = { row = world.rows - 1, col = 0 } 
			corners.ur = { row = 0, col = world.columns - 1 }
			corners.br = { row = world.rows - 1, col = world.columns - 1 } -- bottom right corner
		
		
			for k, v in pairs(corners) do
			
				local lengthx = math.abs(p_col - v.col) -- we want the farthest corner from PLAYER to run away to
				local lengthy = math.abs(p_row - v.row)
				local distance = math.sqrt((lengthx*lengthx)+(lengthy*lengthy))

				if distance > destination.distance then
					destination.row = v.row
					destination.col = v.col
					destination.distance = distance
				end
			end
			
			local destx, desty = world:getTileXY(destination.row, destination.col) 
			
			self.goal = { x = destx, y = desty } -- keep searching for new destiation because we are trying to escape the player.
			--self.needsGoal = false
			
		--end
		
		
		local newpath = game:pathFind(s_row, s_col, self.goal.x, self.goal.y) 
		
		
		
		if newpath ~= nil then
			self.path = newpath
			--print ('set new path')
		else
			--print('newpath is nil')
		end
		
	
		
		
		-- set direction of movement. forget acceleration, friction for now.
		if self.path ~= nil then
			local firstnode
			local nodecount = 0
			
			for node, count in self.path:nodes() do -- why do i have to iterate i just want the second one?? no, just use self.path[i], and self.path:getLength() might be a thing. definitely #self.path works.
				if count == 2 then -- starts at 1
					firstnode = node
				end
				
				nodecount = nodecount + 1
			end

			--print (#self.path)
			
			if nodecount >= 2 then
				local nodex, nodey = world:getTileXY(firstnode.y, firstnode.x) 					
				local newangle = math.angle(sx, sy, nodex, nodey)
				self.angle = newangle
				
				local dest = self.path[#self.path] -- last tile in path
				self:updateFleeingMovement(sh, dt, self.goal.x, self.goal.y) -- destx, desty should be the last node?
			end
		end		
		
	
		
		-- TRY TO FIRE WEAPON 
		self.weapon:update(dt) 
		self:updateAttackingLogic(sh, dt)
 
 end
 
 
 function Boss1:updateFleeingMovement(sh, dt, destx, desty)
	local sx, sy = sh:center()	
	
	local dist = math.dist(sx,sy,destx,desty) 
	
	-- acceleration
	-- maybe stop accelerating when the shortest angle required to turn is >= 90deg? maybe not needed.
	-- stop accelerating as you approach a reasonable range from target (player)
	
	
	if (dist > self.max_fire_range / 3) and self.state ~= "idle" then 
		self.accelerationX = self.accelerationX  + self.thrust * math.cos(self.angle);
		self.accelerationY = self.accelerationY + self.thrust * math.sin(self.angle);
		-- print ('accelerating!')
	else
		self.needsGoal = true -- get new desitnation point for pathfinder.
		-- print ('stopped accelerating')
	end

	-- velocity
	self.speedX = self.speedX + self.accelerationX * dt
	self.speedY = self.speedY + self.accelerationY * dt

	--friction
	self.speedX = self.speedX * self.friction
	self.speedY = self.speedY * self.friction
	
	-- zero out every frame.
	self.accelerationX = 0
	self.accelerationY = 0
	
	if self.speedY > self.maxspeed then
		self.speedY = self.maxspeed
	elseif self.speedY < (self.maxspeed * -1) then
		self.speedY = (self.maxspeed * -1)
	end
	if self.speedX > self.maxspeed then
		self.speedX = self.maxspeed
	elseif self.speedX < (self.maxspeed * -1) then
		self.speedX = (self.maxspeed * -1)
	end
	
	sh:move(self.speedX * dt, self.speedY * dt)

end
  
 function Boss1:updateAttackingState(sh, dt)
	local px, py = player1:center()
	local sx, sy = sh:center()			

	local LOS = game:checkLOS(sh, player1) -- check if we have line of sight to the player. only walls are considered.
	
	if LOS then -- move towards player. otherwise pathfind.
	
		self:updateAngle(sx, sy, px, py, dt)
		self:updateMovement(sh, dt)
	
	else -- NO LOS
	
		-- PATHFIND!
		local row, col = world:getTilePosition(sx, sy, true) -- get tile position in grid
		local newpath = game:pathFind(row, col, px, py) -- TEMP TEMP TESTING
		if newpath ~= nil then
			self.path = newpath
			--print ('set new path')
		else
			--print('newpath is nil')
		end
		
		-- set direction of movement. forget acceleration, friction for now.
		if self.path ~= nil then
			local firstnode
			local nodecount = 0
			
			for node, count in self.path:nodes() do -- why do i have to iterate i just want the second one??
				if count == 2 then -- starts at 1
					firstnode = node
				end
				
				nodecount = nodecount + 1
			end
			
			if nodecount >= 2 then
				local nodex, nodey = world:getTileXY(firstnode.y, firstnode.x) 					
				local newangle = math.angle(sx, sy, nodex, nodey)
				self.angle = newangle
				self:updateMovement(sh, dt)
			end
		end		
	
	end -- end if(LOS)
		
	-- TRY TO FIRE WEAPON 
	self.weapon:update(dt) 
	self:updateAttackingLogic(sh, dt)
 
 end
 
 function Boss1:updateAttackingLogic(sh, dt)
 	local px, py = player1:center()
	local sx, sy = sh:center()		 
	local dist = math.dist(sx,sy,px,py) 		
	
	if self.health > 0 then -- enemy shouldn't attack after it's dead (self.health > 0)
		if (dist < self.max_fire_range) and (dist > self.max_fire_range / 2) then
			
			if self.weapon.weapon_type ~= "spreadbomb" then
				self.weapon = self.weaponSpreadBomb
			end
			
			self:attack(sh) 				
		
		elseif (dist < self.max_fire_range / 2) then
		
			if self.weapon.weapon_type ~= "bomb" then
				self.weapon = self.weaponBomb
			end
			
			self:attack(sh) 				
		
		end
	end
 
 end
  
function Boss1:updateMovement(sh, dt)
	local px, py = player1:center()
	local sx, sy = sh:center()	
	
	local dist = math.dist(sx,sy,px,py) 
	
	-- acceleration
	-- maybe stop accelerating when the shortest angle required to turn is >= 90deg? maybe not needed.
	-- stop accelerating as you approach a reasonable range from target (player)
	
	
	if (dist > self.max_fire_range / 2) and self.state ~= "idle" then 
		self.accelerationX = self.accelerationX  + self.thrust * math.cos(self.angle);
		self.accelerationY = self.accelerationY + self.thrust * math.sin(self.angle);
		-- print ('accelerating!')
	else
		-- print ('stopped accelerating')
	end

	-- velocity
	self.speedX = self.speedX + self.accelerationX * dt
	self.speedY = self.speedY + self.accelerationY * dt

	--friction
	self.speedX = self.speedX * self.friction
	self.speedY = self.speedY * self.friction
	
	-- zero out every frame.
	self.accelerationX = 0
	self.accelerationY = 0
	
	if self.speedY > self.maxspeed then
		self.speedY = self.maxspeed
	elseif self.speedY < (self.maxspeed * -1) then
		self.speedY = (self.maxspeed * -1)
	end
	if self.speedX > self.maxspeed then
		self.speedX = self.maxspeed
	elseif self.speedX < (self.maxspeed * -1) then
		self.speedX = (self.maxspeed * -1)
	end
	
	sh:move(self.speedX * dt, self.speedY * dt)

end
  
function Boss1:setColor()  

	local color = self:getColor()

	love.graphics.setColor(color.r, color.g, color.b)

end

function Boss1:getColor()
	local color = {}

	if self.state == "angry" then
		color.r = 255
		color.g = 50
		color.b = 50
	elseif self.state == "idle" then
		color.r = 75
		color.g = 75
		color.b = 255		
	elseif self.state == "fleeing" then
		color.r = 75
		color.g = 255
		color.b = 75		
	end
	
	return color
end
  
function Boss1:getHit(sh, damage) 
	self.health = self.health - damage
	if self.health >= self.maxhealth / 4 then
		self.state = "angry"
		self.angrytimer = 0
	else
		self.state = "fleeing"
	end
	
	self.spawntimermax = self.spawntimerreset -- reset spawning rate to default when hit.
end 
  
 function Boss1:explode(sh)  --  shape	
	self.health = 0 -- ugh, he will explode in die() function.	
end
  
  
  
 function Boss1:die(i, sh) -- index, shape
 
	player1.instance.score = player1.instance.score + self.pointvalue
 
	local x, y = sh:center()
	local size = self.h
	
	-- explode
	game:spawnExplosion(x, y, size, true)

	-- announce
	game:spawnAnnouncement(x - size / 2, y - size *2, "gold", tostring(self.pointvalue) .. " pts")
	
	
	
	-- remove enemy from table
	Collider:remove(sh)
	table.remove(entities, i)

	-- enemy drops a modifier pickup
	for i = 0, 10 do
		local rx, ry = x + math.random(-100, 100), y + math.random(-100, 100)
		local newEntity = Entities("pickup", rx, ry)
		newEntity:spawn()
	end
		
	
 
 end
 
 function Boss1:resolveCollision(sh, dx, dy)

	-- move shape to resolve collision
	sh:move(dx, dy)
	
	-- rebound 
	if (math.abs(dx) > math.abs(dy)) then
		-- if you are moving in a direction, being hit in the same direction should not cause you to reverse direction
		if dx < 0 and self.speedX > 0 or dx > 0 and self.speedX < 0 then 
			self.speedX = -self.speedX
		end
	else
		if dy < 0 and self.speedY > 0 or dy > 0 and self.speedY < 0 then
			self.speedY = -self.speedY
		end
	end

end

function Boss1:resolveRepel(sh, dx, dy, repel_shape) --
local force = repel_shape.instance.force
local sx, sy = sh:center()
local rx, ry = repel_shape:center()
local repel_angle = math.angle(rx, ry, sx, sy)

	-- move shape to resolve collision
	sh:move(dx, dy) 

	-- rebound 
	if (math.abs(dx) > math.abs(dy)) then
		-- if you are moving in a direction, being hit in the same direction should not cause you to reverse direction
		if dx < 0 and self.speedX > 0 or dx > 0 and self.speedX < 0 then 
			self.speedX = -self.speedX
		end
	else
		if dy < 0 and self.speedY > 0 or dy > 0 and self.speedY < 0 then
			self.speedY = -self.speedY
		end
	end
	
	
	self.speedX = self.speedX + force * math.cos(repel_angle)
	self.speedY = self.speedY + force * math.sin(repel_angle)

end


function Boss1:attack(sh)

	
	local px, py = player1:center()
	local sx, sy = sh:center()
	
	local angle = math.angle(sx, sy, px, py)
	
	local points = {}
	points[0] = { x = sx, y = sy }
	self.weapon:fire(points, self.speedX, self.speedY, angle, self.name) -- FIRE
	
	
	
end
  
  
   
return Boss1