local class = require 'lib.middleclass'

Bullet = class('Bullet')

function Bullet:initialize(_x, _y, _speedx, _speedy, _angle, origin, modifiers) --originating entity
	
	self.id = math.random(1,100000) -- for testing purposes
	self.h = 5
	self.w = 5
	self.radius = 3
	self.x = _x
	self.y = _y
	self.angle = _angle
	self.firingspeed = 500
	
	self.lifespan = 1.2
	self.damage = 25
	self.firingtimer = .2
	self.bounces = 0
	self.origin = origin --originating entity (player/enemy)
	
	prop_max = WeaponModifiers:getBulletPropMax()
	
	if (modifiers ~= nil) then -- not sure i need to check 
		self.firingspeed = self.firingspeed + modifiers.fs
		self.lifespan = self.lifespan + modifiers.ls
		self.damage = self.damage + modifiers.dmg
		self.firingtimer = self.firingtimer - modifiers.ft
		self.bounces =  self.bounces + modifiers.b
	end
	
	
	self.speedX = _speedx + self.firingspeed  * math.cos(_angle)
	self.speedY = _speedy + self.firingspeed  * math.sin(_angle)
	
end

function Bullet:draw(sh)
	
	love.graphics.setColor(255,255,255)		
	sh:draw('fill')
	
end

function Bullet:update(sh, dt)
			
	local sx, sy = sh:center()
	local s = self
	
	sx = sx + s.speedX * dt
	sy = sy + s.speedY * dt
	
	s.lifespan = s.lifespan - dt	
	
	sh:move(s.speedX * dt, s.speedY * dt)
		
end


function Bullet:getHit(sh, damage)
	self:explode(sh)
end

function Bullet:explode(sh)  --  shape	
	
	local sx, sy = sh:center()
	game:spawnExplosion(sx, sy, 40) 
	self.lifespan = 0
	
end

function Bullet:die(i, sh) -- we need index to remove it from bullet table.
									-- if you want a bullet to die outside of main.lua update bullets loop, use self.lifespan=0

	Collider:remove(sh)
	table.remove(bullets, i)
	--print ('removed bullet ' .. i )
	
end

function Bullet:resolveCollision(sh, dx, dy)
	
	-- move shape to resolve collision
	sh:move(dx, dy) 
		
	if self.bounces > 0 then
		-- reverse speed on the axis with the largest distance to travel to resolve the collision
		-- creates a rebound
		if (math.abs(dx) > math.abs(dy)) then
			self.speedX = -self.speedX
		else 
			self.speedY = -self.speedY
		end
		self.bounces = self.bounces - 1
	else
		self.lifespan = 0
	end
		
end


function Bullet:resolveRepel(sh, dx, dy, repel_shape) --
local force = repel_shape.instance.force
local sx, sy = sh:center()
local rx, ry = repel_shape:center()
local repel_angle = math.angle(rx, ry, sx, sy)

	-- move shape to resolve collision
	sh:move(dx, dy) 

	-- rebound 
	if (math.abs(dx) > math.abs(dy)) then
		-- if you are moving in a direction, being hit in the same direction should not cause you to reverse direction
		if dx < 0 and self.speedX > 0 or dx > 0 and self.speedX < 0 then 
			self.speedX = -self.speedX
		end
	else
		if dy < 0 and self.speedY > 0 or dy > 0 and self.speedY < 0 then
			self.speedY = -self.speedY
		end
	end
	
	
	self.speedX = self.speedX + force * math.cos(repel_angle)
	self.speedY = self.speedY + force * math.sin(repel_angle)

end


function Bullet:playSound()
	if sfx.bullet:isPlaying() then
		local newSource = sfx.bullet:clone()
		newSource:play()
	else
		sfx.bullet:play()
	end
end

return Bullet