local class = require 'lib.middleclass'


Enemy3 = class('Enemy3')

function Enemy3:initialize(x, y, _size) 

	self.name = "enemy"
	
	self.id = math.random(1,100000) -- for testing purposes
		
	self.points = 
				{
					x1 = 60,
					y1 = 20,
					x2 = 100,
					y2 = 80,
					x3 = 60,
					y3 = 140,
					x4 = 20,
					y4 = 80
				 }
	
	self.h = 60 -- useful for healthbar & title positioning
	self.w = 30 -- not sure this will be used.
	self.x = x
	self.y = y
	self.angle = 0
	self.firingspeed = 0
	self.rotatespeed = 2.5
	self.accelerationX = 0
	self.accelerationY = 0
	self.friction = 0.99
	self.thrust = 0 --math.random(300,600)
	self.maxspeed = 500 -- math.random(500,1200)
	self.speedX = 100 --math.random(-350, 350)
	self.speedY = 100 --math.random(-350, 350)
	
	local modifier = game:getStageModifier() 
	
	-- this enemy doesn't split atm.
	self.maxhealth = 500 * modifier -- add health based on stage modifier (and for enemy2, spawn size)
	self.health = self.maxhealth
	
	self.damagevalue = 10 -- normally this would be a weapon or projectile property.
	
	self.pointvalue = math.round(1000 / self.h) * modifier
	
	self.min_fire_range = 0
	self.max_fire_range = 800
	
	
	self.lockedon = false
	self.sounds = sfx.raygun:clone()
	
end

function Enemy3:draw(sh)
		
	
	
	local px, py = player1:center()
	local sx, sy = sh:center()
	self:setColor()
	local r,g,b = love.graphics.getColor()
	
	love.graphics.setColor(r,g,b,15) 

	-- draw range radius
	local thickness = self.max_fire_range - self.min_fire_range
	love.graphics.setLineWidth( thickness )
	love.graphics.circle('line', sx, sy, self.min_fire_range + (thickness/2))
	
	-- draw shape (self)
	love.graphics.setLineWidth(8)
	self:setColor()	 --love.graphics.setColor(r,g,b,255)
	sh:draw('line')
	love.graphics.setColor(255,255,255)
	love.graphics.circle('fill',sx, sy,10)
	love.graphics.setLineWidth(1)			
		
	self:setColor()
	-- DRAW HEALTH BAR
	local percent = (self.health / self.maxhealth) * 100
	local lx, ly = sx - self.h, sy - self.h - 20
	local lx2, ly2 = sx + self.h, ly
	local dist = math.dist(lx, ly, lx2, ly2)
	local line_dist = dist * (percent/100)
	love.graphics.line(lx, ly, lx + line_dist, ly2)
	
	-- PRINT SELECTED WEAPON
	love.graphics.setColor(255,255,255)
	local font_w = mainFont:getWidth("BEAM") -- Even though it's not a "beam" per se. At least, it is not a entity class beam.
	love.graphics.print("BEAM", sx - font_w , ly - 50, 0, 2, 2)
	
	
	-- DRAW WEAPON (beam type thing, but not projectile / not indexed in bullet table).
	if self.lockedon then
		
		local dist = math.dist(sx, sy, px, py)
		dist = dist - player1.instance.ship.shield.radius
		local angle = math.angle(sx, sy, px, py)
		local newx = sx + dist * math.cos(angle) 
		local newy = sy + dist * math.sin(angle)
		
		
		
		if  player1.instance.ship.shield.state == 'on' then
			love.graphics.setLineWidth(4)
			love.graphics.line(sx, sy, newx, newy)
			love.graphics.circle('fill',sx, sy,10)
			love.graphics.setLineWidth(1)		
		else
			love.graphics.setColor(255,0,0)
			love.graphics.setLineWidth(8)
			love.graphics.line(sx, sy, px, py)
			love.graphics.circle('fill',sx, sy,10)
			love.graphics.setLineWidth(1)				
		end
		
	end
	
end

function Enemy3:update(sh, dt)
		local px, py = player1:center()
		local sx, sy = sh:center()			

		self:updateMovement(sh, dt)

		local LOS = game:checkLOS(sh, player1) -- check if we have line of sight to the player. only walls are considered.
			
		-- TRY TO FIRE WEAPON 
		
		local dist = math.dist(sx,sy,px,py) 		
		if (dist < self.max_fire_range) and (dist > self.min_fire_range) then
			if LOS then
				self.lockedon = true
				
				if player1.instance.ship.shield.state == 'off' then
					if self.health > 0 then -- enemy shouldn't attack after it's dead (self.health > 0)
						self:attack(sh) 				
					end
				end
				
			else
				if self.lockedon then
					self.lockedon = false
				end
			end
		
		else
			self.lockedon = false
		end
	
end
  
function Enemy3:updateMovement(sh, dt)
	local px, py = player1:center()
	local sx, sy = sh:center()	
	
	
	--friction
	self.speedX = self.speedX * self.friction
	self.speedY = self.speedY * self.friction
	
	if self.speedY > self.maxspeed then
		self.speedY = self.maxspeed
	elseif self.speedY < (self.maxspeed * -1) then
		self.speedY = (self.maxspeed * -1)
	end
	if self.speedX > self.maxspeed then
		self.speedX = self.maxspeed
	elseif self.speedX < (self.maxspeed * -1) then
		self.speedX = (self.maxspeed * -1)
	end
	
	sh:move(self.speedX * dt, self.speedY * dt)

end
  
function Enemy3:setColor()  

	local color = self:getColor()

	love.graphics.setColor(color.r, color.g, color.b, 255)

end

function Enemy3:getColor()
	local color = {}

	color.r = 0
	color.g = 100
	color.b = 0

	return color
end
  
function Enemy3:getHit(sh, damage) 
	self.health = self.health - damage
end 
  
 function Enemy3:explode(sh)  --  shape	
	self.health = 0 -- ugh, he will explode in die() function.	
end
  
  
  
 function Enemy3:die(i, sh) -- index, shape
 
	player1.instance.score = player1.instance.score + self.pointvalue
 
	local x, y = sh:center()
	local size = self.h
	
	-- explode
	game:spawnExplosion(x, y, size, true)

	-- announce
	game:spawnAnnouncement(x - size / 2, y - size *2, "gold", tostring(self.pointvalue) .. " pts")
	
	
	
	-- remove enemy from table
	Collider:remove(sh)
	table.remove(entities, i)

	-- enemy drops a modifier pickup
	local newEntity = Entities("pickup", x, y)
	newEntity:spawn()
		
	
 
 end
 
 function Enemy3:resolveCollision(sh, dx, dy)

	-- move shape to resolve collision
	sh:move(dx, dy)
	
	-- rebound 
	if (math.abs(dx) > math.abs(dy)) then
		-- if you are moving in a direction, being hit in the same direction should not cause you to reverse direction
		if dx < 0 and self.speedX > 0 or dx > 0 and self.speedX < 0 then 
			self.speedX = -self.speedX
		end
	else
		if dy < 0 and self.speedY > 0 or dy > 0 and self.speedY < 0 then
			self.speedY = -self.speedY
		end
	end

end

function Enemy3:resolveRepel(sh, dx, dy, repel_shape) --
local force = repel_shape.instance.force
local sx, sy = sh:center()
local rx, ry = repel_shape:center()
local repel_angle = math.angle(rx, ry, sx, sy)

	-- move shape to resolve collision
	sh:move(dx, dy) 

	-- rebound 
	if (math.abs(dx) > math.abs(dy)) then
		-- if you are moving in a direction, being hit in the same direction should not cause you to reverse direction
		if dx < 0 and self.speedX > 0 or dx > 0 and self.speedX < 0 then 
			self.speedX = -self.speedX
		end
	else
		if dy < 0 and self.speedY > 0 or dy > 0 and self.speedY < 0 then
			self.speedY = -self.speedY
		end
	end
	
	
	self.speedX = self.speedX + force * math.cos(repel_angle)
	self.speedY = self.speedY + force * math.sin(repel_angle)

end


function Enemy3:attack(sh)
	local px, py = player1:center()
	local sx, sy = sh:center()
	
	if self.lockedon then
		if not self.sounds:isPlaying() then
			self.sounds:play()
		end
		
		--just dole out the dmg randomly atm.
		if math.random(1,100) > 99 then
			player1.instance:takeDamage(self.damagevalue) 
			--print(tostring(self.id) .. ' - dealt damage : ' .. tostring(self.damagevalue))
		end		
		
	else
		if self.sounds:isPlaying() then
			self.sounds:stop()
			self.sounds:rewind()
		end
	end
		
end
  
  
   
return Enemy3