
local class = require 'lib.middleclass'

Entities = class('Entities')

function Entities:initialize(entity_type, x, y, size, speedx, speedy, angle)

	self.entity_type = entity_type
	self.x = x
	self.y = y
	self.size = size or nil
	self.speedx = speedx or 0
	self.speedy = speedy or 0
	self.angle = angle or 0

end

function Entities:spawn()
local entity, shape

--print(self.entity_type)

	if (self.entity_type == 'enemy1') then			
		
		entity = Enemy1(self.x, self.y, self.size) -- 3rd param size is optional
		shape = Collider:addRectangle(entity.x, entity.y, entity.w, entity.h)
		shape.type = "enemy" -- "enemy1"
		shape.name = "enemy1" --  "enemy1"
		
		-- for fun, 25% chance enemy has double HP but starts at 1/2 hp, so it can shoot after spawning.
		if math.random(1,100) > 75 then
			entity.maxhealth = entity.maxhealth * 2
			entity.weapon.armed = true
			--print(entity.health, entity.maxhealth)
		end
	elseif (self.entity_type == 'enemy2') then
	
		entity = Enemy2(self.x, self.y, self.size) -- 3rd param size is optional
		shape = Collider:addCircle(entity.x, entity.y, entity.h)
		shape.type = "enemy" --  "enemy2"
		shape.name =  "enemy2" -- "enemy2"
	
	elseif (self.entity_type == 'pickup') then
		
		entity = Pickup(self.x, self.y)
		shape = Collider:addRectangle(entity.x, entity.y, entity.w, entity.h)
		shape.type = "pickup"
		shape.name = "pickup"
		
	elseif (self.entity_type == "boss1") then
	
		entity = Boss1(self.x, self.y) -- 3rd param size is optional
		shape = Collider:addCircle(entity.x, entity.y, entity.h)
		shape.type = "enemy" 
		shape.name =  "boss1"
	
	elseif (self.entity_type =="enemy3") then
	
		entity = Enemy3(self.x, self.y) -- 3rd param size is optional
		local pts = entity.points
		shape = Collider:addPolygon(pts.x1, pts.y1, pts.x2, pts.y2, pts.x3, pts.y3, pts.x4, pts.y4)
		shape:moveTo(entity.x, entity.y)
		shape.type = "enemy" 
		shape.name =  "enemy3"
	
	end
	
	
	shape.instance = entity
	table.insert(entities, shape)	-- add to entities table
			
end

function Entities:update(dt) -- i think i can call this without instantiating as long as i dont use any class properties? yes

local wl, wt, ww, wh = game.cam1:getWorld()
local x, y = wl + world.bufferpixels/2, wt + world.bufferpixels/2
local x2, y2 = x + (ww - world.bufferpixels), y + (wh - world.bufferpixels)
local border_width = world.borderwidth

	--update entities
	for i, v in ipairs(entities) do
		v.instance:update(v, dt)
		
		-- check if entity has left the zone borders, and destroy it
		if v.type == "enemy" then
			local sx, sy = v:center()
			if not point_contained(x - border_width, y - border_width, x2 + border_width, y2 + border_width, sx, sy) then
				v.instance:die(i, v)
			end
		end

			
		if (v.instance.health <= 0) then -- using this for pickups. on collision, set health to 0, and pickup "dies".
			v.instance:die(i, v)			
		end
	end

end


function Entities:draw()
	
	-- debugging destroy enemies leaving world borders
	-- local wl, wt, ww, wh = game.cam1:getWorld()
	-- local x, y = wl + world.bufferpixels/2, wt + world.bufferpixels/2
	-- local border_width = world.borderwidth
	-- love.graphics.setColor(255,0,0)
	-- love.graphics.rectangle('line', x - border_width, y - border_width, ww - world.bufferpixels + border_width*2, wh - world.bufferpixels + border_width*2)

	
	for i, v in ipairs(entities) do
		
		local e_size = v.instance.h -- i would do entity_size / 2 but my circle entities use this for radius. also i use this value to position the entity's health bar & weapon name.
		
		local l,t,w,h = game.cam1:getVisible()
		local x1, y1 = l - e_size, t - e_size
		local x2, y2 = l + w + e_size, t + h + e_size
		
		local entity_x, entity_y = v:center()
		
		if point_contained(x1, y1, x2, y2, entity_x, entity_y) then
			
			v.instance:draw(v)

		else
			--not in camera
			self:drawInRadar(entity_x, entity_y, v.instance:getColor())
			
		end				
	end

end

function Entities:drawInRadar(entity_x, entity_y, color)

			local l,t,w,h = game.cam1:getVisible()
			local px, py = player1:center()
			--local angle_from_player = math.angle(px, py, enemy_x, enemy_y)
			
			local dist = math.dist(px, py, entity_x, entity_y)
			local icon_size = ((world.w * 5)/dist)  -- icon grows as it gets closer
			if icon_size < 8 then icon_size = 8 end
			
			-- N88 edge screen radar 
			love.graphics.setColor(color.r, color.g, color.b)
			
			love.graphics.push()
			love.graphics.translate(l,t)
										
			if entity_x < px then
				local hor = (entity_x-px)/(w/2)
				local vert = (entity_y-py)/(h/2)
				love.graphics.circle("fill",0+20,(h/2)-(entity_y-py)/hor, icon_size)
			end
			if entity_x > px then
				local hor = (entity_x-px)/(w/2)
				local vert = (entity_y-py)/(h/2)
				love.graphics.circle("fill",w-20,(h/2)+(entity_y-py)/hor, icon_size)
			end
			if entity_y < py then
				local hor = (entity_x-px)/(w/2)
				local vert = (entity_y-py)/(h/2)
				love.graphics.circle("fill",(w/2)-(entity_x-px)/vert,0+20, icon_size)
			end
			if entity_y > py then
				local hor = (entity_x-px)/(w/2)
				local vert = (entity_y-py)/(h/2)
				love.graphics.circle("fill",(w/2)+(entity_x-px)/vert,h-20, icon_size)
			end
			
			love.graphics.pop()		

end

function Entities:resetAll()
	for i, v in ipairs(entities) do		
		Collider:remove(v)
	end
	entities = {}
end


function Entities:reset(entity_type)
	for i, v in ipairs(entities) do
		if (v.type == entity_type) then				
			Collider:remove(v)
			table.remove(entities, i)
		end		
	end
end

return Entities