-- Maze generator in Lua
-- Joe Wingbermuehle 2013-10-06

local class = require 'lib.middleclass'



Levels = class('Levels')

function Levels:initialize(rows, cols) -- rows cols must be ODD!

	self.rows = rows
	self.cols = cols
	
	--rooms
	self.maxrooms = 1000
	self.minsize = 2
	self.maxsize = 4
	
	
	
end

function Levels:generateCustom(map_tiles)
	local finaltiles = {}
		
	--create table....
	for y = 0, self.rows - 1 do
		finaltiles[y] = {}
		for x = 0, self.cols - 1 do
			finaltiles[y][x] = map_tiles[y+1][x+1]
		end
	end

	
	return finaltiles
end

function Levels:generateEmpty()
	local finaltiles = {}
		
	--create table....
	for y = 0, self.rows - 1 do
		finaltiles[y] = {}
		for x = 0, self.cols - 1 do
			finaltiles[y][x] = ' '
		end
	end
	
	return finaltiles
end


function Levels:generate()
	-- The size of the maze (must be odd).
	local width = self.cols
	local height = self.rows

	-- Generate and display a random maze.
	local maze = self:init_maze(width, height)
	maze = self:carve_maze(maze, width, height, 2, 2)
	maze[width + 2] = 0 --entry/exit
	maze[(height - 2) * width + width - 3] = 0 --entry/exit
	
	--self:show_maze(maze, width, height) -- for testing
	
	local roomtiles = self:generate_room_tiles() -- this wallgen doesnt work great even with higher tile count
	local finaltiles = self:add_rooms(maze, roomtiles)
	
	return finaltiles
end




-- Create an empty maze array.
function Levels:init_maze(width, height)
   local result = {}
   for y = 0, height - 1 do
      for x = 0, width - 1 do
         result[y * width + x] = 1
      end
      result[y * width + 0] = 0
      result[y * width + width - 1] = 0
   end
   for x = 0, width - 1 do
      result[0 * width + x] = 0
      result[(height - 1) * width + x] = 0
   end
   return result
end

-- Show a maze.
function Levels:show_maze(maze, width, height)
   for y = 0, height - 1 do
      for x = 0, width - 1 do
         if maze[y * width + x] == 0 then -- 0 == space, 1 == wall
            io.write(" ")
         else
            io.write("#")
         end
      end
      io.write("\n")
   end
end

-- Carve the maze starting at x, y.
function Levels:carve_maze(maze, width, height, x, y)
   local r = math.random(0, 3)
   maze[y * width + x] = 0
   for i = 0, 3 do
      local d = (i + r) % 4
      local dx = 0
      local dy = 0
      if d == 0 then
         dx = 1
      elseif d == 1 then
         dx = -1
      elseif d == 2 then
         dy = 1
      else
         dy = -1
      end
      local nx = x + dx
      local ny = y + dy
      local nx2 = nx + dx
      local ny2 = ny + dy
      if maze[ny * width + nx] == 1 then
         if maze[ny2 * width + nx2] == 1 then
            maze[ny * width + nx] = 0
            self:carve_maze(maze, width, height, nx2, ny2)
         end
      end
   end
   
   return maze
end


-- ####################################################### --


function Levels:generate_room_tiles()

	local mapw, maph, maprows, mapcols = world.w, world.h, self.rows, self.cols 

	local w = math.round(mapw / mapcols - 1) -- 1 column width
	local h = math.round(maph / maprows - 1) -- 1 row width
	--local w = 10
	--local h = 5

	--print (w, h)
	local rooms = {}
	local tile = {}

	for i=1, maprows do
		tile[i] = {}
		for j=1, mapcols do
			tile[i][j] = '#' -- set everything to wall
		end
	end

	local failed_add = false
	for i=0, self.maxrooms do

		local w = math.random(self.minsize, self.maxsize)
		local h = math.random(self.minsize, self.maxsize)
		local x1 = math.random(1, maprows - h) -- x position in the table, not pixels
		local y1 = math.random(1, mapcols - w) -- y position in table

		local newroom = {
							w = w,
							h = h,
							x1 = x1, -- these are actually ROW positions. (technically it's like y pos)
							y1 = y1, -- these are actually COL positions. (technically it's like x pos) confusing...
							x2 = x1 + h,
							y2 = y1 + w
						}

		for i, v in ipairs(rooms) do
			if intersects(newroom, v) then
				failed_add = true
			end
		end

		if not failed_add then
			table.insert(rooms, newroom)
		end
	end

	--print('good rooms below')

	-- add rooms to tile array as spaces
	for k, v in ipairs(rooms) do
		--print(v.x1, v.y1, v.x2, v.y2)
		--cycle through 2d array and swap wall for space
		for i, row in ipairs(tile) do
			for j, col in ipairs(tile [i]) do
				if ( i >= v.x1 and i <= v.x2 ) and (j >= v.y1 and j <= v.y2) then
					tile[i][j] = " "
				end
			end
		end
	end

	
	-- print tiles with generated rooms
	for i, row in ipairs(tile) do
		for j, col in ipairs (tile[i]) do
			--io.write (tile[i][j])
		end
		--io.write('\n');
	end
	--io.write("\n")

	
	
	return tile
end

function Levels:add_rooms(maze, roomtiles) -- maze tiles, room tiles

	local finaltiles = {}
	local count = 0 -- debug
	local maprows = self.rows
	local mapcols = self.cols

	--create table....
	for y = 0, maprows do
		finaltiles[y] = {}
		for x = 0, mapcols do
			finaltiles[y][x] = '^' -- set everything to whatever
		end
	end
	
	--io.write("\n")
	-- print table using maze loops
	for y = 1, maprows do
		for x = 1, mapcols do
			--print(y, x)
			
			-- if room tile is a space
			if (roomtiles[y][x] == " ") then
				-- print rooms table if it's a empty space
				--io.write(roomtiles[y][x])
				
				finaltiles[y-1][x-1] = roomtiles[y][x]
			
			-- if it's not a space it's a wall. check if equivalent maze tile is a space
			elseif (maze[(y-1) * (mapcols) + (x-1)] == 0) then --  have to -1 because the maze generator starts at 0
			
				--io.write(" ") 
				finaltiles[y-1][x-1] = " " -- use a space if it is.
			else
				--io.write("#") 
				finaltiles[y-1][x-1] = "#" -- otherwise it's a wall
				count = count + 1
			end
		end
		--io.write("\n")
	end


	--io.write("\n")
	--print("wall count", count);
	
	return finaltiles
end


return Levels