local class = require ('lib.middleclass')
local gamera = require('lib.gamera')
HC = require ('lib.hardoncollider')
jumper_grid = require 'lib.jumper181.jumper.grid'
jumper_pf = require 'lib.jumper181.jumper.pathfinder'


require 'util.helper' 
require 'game'
require 'world'
require 'player'
require 'explosion'
require 'enemy1'
require 'enemy2'
require 'collisions'
require 'weapons'
require 'bullet'
require 'bomb'
require 'repel'
require 'levels'
require 'entities'
require 'pickup'
require 'announce'
require 'util.controls'
require 'weaponModifiers'
require 'boss1'
require 'enemy3'


math.randomseed(os.time())
game = Game() -- super global game instance


-- camera functions
local function updateCameras(dt)
  
	--local p = player1.ship -- remove this.
	local px, py = player1:center()

	game.cam1:setPosition(px, py)
	--game.cam2:setPosition(px, py)

	-- zoom camera
  
	local scaleFactor = 0
	
	-- REMOVED, for debug only.
	if game.settings.debug then 
		scaleFactor = isDown('s') and -0.8 or (isDown('w') and 0.8 or 0)
	end

	game.cam1:setScale(game.cam1:getScale() + scaleFactor * dt)
	
end

local function drawCam1ViewPort()
  
  love.graphics.setColor(37,82,86, 150)
  love.graphics.rectangle('fill', game.cam1:getVisible())
  --love.graphics.polygon('line', game.cam1:getVisibleCorners()) -- this is for rotation on the minimap.
end




-- BASE LOVE FUNCTIONS
function love.load()
	-- NOTE VSYNC OFF PREVENTS SHIP FROM REACHING MAXSPEED
	love.window.setMode(1280, 720 , {resizable=true, vsync=true, minwidth=1280, minheight=720})
	
	
	-- SFX
	sfx = {}
	sfx.thruster = love.audio.newSource('sfx/thruster3.mp3', 'static') -- http://soundbible.com/1492-Rocket-Thrusters.html (trimmed)
	sfx.thruster:setLooping(true)
	sfx.thruster:setVolume(0.2)
	
	sfx.bounce = love.audio.newSource('sfx/bounce3.mp3', 'static') -- http://soundbible.com/1120-Bounce.html (altered)
	sfx.bounce:setLooping(false)
	sfx.bounce:setVolume(0.5)
	
	sfx.bullet = love.audio.newSource('sfx/pew2.mp3', 'static') -- http://soundbible.com/1949-Pew-Pew.html (trimmed)
	sfx.bullet:setLooping(false)
	sfx.bullet:setVolume(0.1)
	
	sfx.explosion = love.audio.newSource('sfx/explosion2.mp3', 'static') -- http://soundbible.com/763-Bomb.html (altered)
	sfx.explosion:setLooping(false)
	sfx.explosion:setVolume(0.6)
	
	sfx.siren = love.audio.newSource('sfx/siren2.mp3', 'static') -- http://soundbible.com/1937-Tornado-Siren-II.html (altered)
	sfx.siren:setLooping(false)
	sfx.siren:setVolume(0.1)
	
	sfx.door_slam = love.audio.newSource('sfx/door_slam2.mp3', 'static') -- http://soundbible.com/1551-Large-Door-Slam.html (amplified/trimmed)
	sfx.door_slam:setLooping(false)
	sfx.door_slam:setVolume(1)
	
	sfx.ding = love.audio.newSource('sfx/ding2.mp3','static')  -- http://soundbible.com/1441-Elevator-Ding.html 
	sfx.ding:setLooping(false)
	sfx.ding:setVolume(0.5)
	
	sfx.cheer = love.audio.newSource('sfx/small_crowd_applause2.mp3', 'static') -- http://soundbible.com/1964-Small-Crowd-Applause.html (trimmed)
	sfx.cheer:setLooping(false)
	sfx.cheer:setVolume(0.7)
	
	sfx.wrench = love.audio.newSource('sfx/air_wrench2.mp3', 'static') -- http://soundbible.com/1975-Air-Wrench-Short.html (trim / double / pitch)
	sfx.wrench:setLooping(false)
	sfx.wrench:setVolume(0.2)
	
	-- I would prefer a more burning/electrical type of sound for this. it's not really firing a laser so much as burning you with a locked on beam.
	sfx.raygun = love.audio.newSource('sfx/raygun.mp3', 'static') -- http://soundbible.com/1770-Ray-Gun.html 
	sfx.raygun:setLooping(false)
	sfx.raygun:setVolume(0.2)
		
	
	-- really great stuff here http://audionautix.com/ 
	sfx.bgm = {}
	sfx.bgm[0] = love.audio.newSource('sfx/bgm/inthefield.mp3', 'stream')
	sfx.bgm[0]:setLooping(true)
	sfx.bgm[0]:setVolume(0.3)
	sfx.bgm[1] = love.audio.newSource('sfx/bgm/alison.mp3', 'stream')
	sfx.bgm[1]:setLooping(true)
	sfx.bgm[1]:setVolume(0.3)
	sfx.bgm[2] = love.audio.newSource('sfx/bgm/dirtroadtraveller.mp3', 'stream')
	sfx.bgm[2]:setLooping(true)
	sfx.bgm[2]:setVolume(0.3)
	sfx.bgm[3] = love.audio.newSource('sfx/bgm/rubixcube.mp3', 'stream')
	sfx.bgm[3]:setLooping(true)
	sfx.bgm[3]:setVolume(0.3)
	
	-- master volume
	love.audio.setVolume(0.4)
	
	-- FONTS
	mainFont = love.graphics.newFont('fonts/prototype.ttf',16) -- http://www.dafont.com/prototype.font
	largeFont = love.graphics.newFont('fonts/prototype.ttf',40)
	giganticFont = love.graphics.newFont('fonts/prototype.ttf',350)

		
	-- gamepad mappings
	-- I don't think i really need this but i'm disabling it for the time being anyway.
	-- love.joystick.loadGamepadMappings('util/gamecontrollerdb.txt')
	
	-- start game
	game:start()

end



function love.update(dt)
	
	if player1.instance.ship.alive == true then
	
		Collider:update(dt) -- check collisions
		player1.instance:update(player1, dt)
		game:update(dt)			
		updateCameras(dt)

	end
	
	
	
	-- MOVED TO CONTROLS.LUA >> didn't move joystick restart though.
	-- check for joystick input. this feels like overkill
	-- local joy1 = inputDevice.joystick ~= nil and inputDevice.joystick or nil
	-- local joy1restart = false
	-- if joy1 ~= nil then
		-- if joy1:isGamepadDown('start') then
			-- joy1restart = true
		-- end
	-- end
	
	-- -- this always needs to be available.
	-- if isDown('escape') or joy1restart then
		-- game:start()
	-- end
	
	
end

function love.draw()


	-- draw main camera
	game.cam1:draw(function(l,t,w,h)
		
		world:drawBackground()
		
		world:drawWalls()
		player1.instance:draw(player1)
		game:drawProjectiles() 
		Entities:draw() -- draw enemies / pickups
		game:drawExplosions()
		game:drawAnnouncements()
	
		
	end)
		
		
	
		
		
	-- camera window borders 
	love.graphics.setColor(79, 124, 176)
	love.graphics.rectangle('line', game.cam1:getWindow())

	
	
	
	
	
	
local msg = [[CONTROLS

arrow keys : move	
spacebar : fire bullet
TAB : fire bomb
shift-TAB : repel
shift: afterburner
ESC : respawn

MISC
camera: 8 or 9 or 0
f : hide this sidebar
z, x : volume down / up
Hold L-ALT : large zone map


DEBUG
2 : toggle debug

F1 : console get tile
5 : pump buffs
1 : spawn pickup
w, s :  camera zoom
]]
	
	-- GET WINDOW BOUNDARIES	
	local window_width, window_height, window_flags = love.window.getMode()
	local sidebar_width = 240
	local cl, ct, cw, ch = game.cam1:getWindow()
	
	local modifier = game:getStageModifier() 
	love.graphics.setColor(255,255,255,255)
	love.graphics.setFont(mainFont) -- SET FONT
	
	if (game.settings.sidebar == true) then
		
		love.graphics.setLineWidth(2)
		love.graphics.line(window_width - sidebar_width, 500, window_width - 20, 500) 
		
		-- SIDE AREA
		love.graphics.print(msg, window_width - sidebar_width, 15)
		
		love.graphics.print('hi score (this session): \n' .. game.hiscore, window_width - sidebar_width, 520)
		love.graphics.print('current score: \n' .. player1.instance.score, window_width - sidebar_width, 560)
		love.graphics.print('stage modifier: \n' ..  modifier, window_width - sidebar_width, 600)
		
	end
	
	
	
	-- ON SCREEN TOP MIDDLE
	
	love.graphics.print("Zone Unlocks In: \n" .. math.round(game.leveltimer), cw / 2 - 100, 15)	-- fix horizontal position later
	
	-- ON SCREEN TOP RIGHT
	
	love.graphics.print('Current Score: \n' .. player1.instance.score, cw - 150 , 15) -- fix horizontal position later
	
	-- ON SCREEN LOWER RIGHT
	--print (player1.instance.xquadrant, player1.instance.yquadrant)
	love.graphics.print("ZONE : " .. player1.instance.xquadrant .. ", " .. player1.instance.yquadrant, cw - 130, ch - 20)-- fix horizontal position later
	
	-- LOWER AREA
	
	local bombtimer = math.abs(math.round(player1.instance.weaponbomb.firingtimer,1)) -- fix horizontal position later
	if (bombtimer <= 0) then
		love.graphics.setColor(255,215,0)
	else
		love.graphics.setColor(178,34,34)
	end
	love.graphics.print("Next Bomb : " .. bombtimer , cw / 2 - 100, ch - 20)
	love.graphics.setColor(255,255,255)
	
	
	love.graphics.setLineWidth(2)

	-- player health meter
	local percent = (player1.instance.health / player1.instance.maxhealth) * 100
	percent = percent >= 0 and percent or 0 -- dont display below 0 health
	local bar_length = 145
	local percent_length = bar_length * (percent/100)
	
	love.graphics.setColor(255, 10, 10)
	love.graphics.rectangle('fill', 15 , ch - 110, percent_length, 10)
	love.graphics.setColor(255,255,255)
	love.graphics.rectangle('line', 15, ch - 110, bar_length, 10)
	love.graphics.print("Health : " .. tostring(player1.instance.health), 15, ch - 130)
	
	
	-- Display bullet/bomb modifier METERS
	-- need to create a table to load prop_max from. so i can reference without an instance. use the table for the projectile entity as well
	
	local prop_max = WeaponModifiers:getBulletPropMax()
	local bomb_prop_max = WeaponModifiers:getBombPropMax()
		
	local count = 1
	for i, v in pairs(prop_max) do  -- draw bullet modifier meters
			
		local mod_min = 0
		local mod_max = v
		local mod_current = player1.instance.weaponguns.modifiers[i]
	
		local percent = (mod_current / mod_max) * 100
		percent = percent <= 100 and percent or 100
		
		love.graphics.setColor(255,255,255)
		love.graphics.rectangle('fill', 15 * count, ch + 5, 10, -percent)
		love.graphics.setColor(255,255,255)
		love.graphics.rectangle('line', 15 * count, ch + 5, 10, -100)
		
		count = count + 1 -- for determining x position of the meter
	end
	
	for i, v in pairs(bomb_prop_max) do  -- draw bomb modifier meters
			
		local mod_min = 0
		local mod_max = v
		local mod_current = player1.instance.weaponbomb.modifiers[i]
	
		local percent = (mod_current / mod_max) * 100
		percent = percent <= 100 and percent or 100
			
		love.graphics.setColor(30,144,255)
		love.graphics.rectangle('fill', 15 * count, ch + 5 , 10, -percent)
		love.graphics.setColor(255,255,255,255)
		love.graphics.rectangle('line', 15 * count, ch + 5, 10, -100)
		
		count = count + 1 -- for determining x position of the meter
	end
	-- END TEMPORARY 
	
	
	-- REPEL STOCKS
	if player1.instance.ship.repel.count > 0 then
		love.graphics.setColor(255,215,0)
	else
		love.graphics.setColor(178,34,34)
	end
	
	for i = 0, player1.instance.ship.repel.count - 1 do
		love.graphics.rectangle('fill', 15 * count, ch - (i*20) - 12, 10, 18)
	end
	
	
	
	
	-- ZONE MINI-MAP
	
	local grid_size = 5 -- must be ODD
	local cell_size = 10
	local center_on_player = true
		
	if love.keyboard.isDown("lalt") then -- show full grid?
		grid_size = 25
		cell_size = 20
		center_on_player = false -- don't center on player for full grid display.
	end
		
	local cell_padding = 2
	local left_start = (cw - (cell_size+cell_padding)) - grid_size * (cell_size + cell_padding)
	local top_start = (ch - 15 - (cell_size+cell_padding)) - grid_size * (cell_size + cell_padding)
		
	-- grid BG
	love.graphics.setColor(0,0,0)
	love.graphics.rectangle('fill', left_start - cell_padding, top_start - cell_padding, grid_size * (cell_size + cell_padding) +cell_padding, grid_size * (cell_size + cell_padding)+cell_padding)

	local grid_offset = math.floor(grid_size/2)
	local current = world.currentstage	
	local y_cell_count = center_on_player and current.y - grid_offset or -grid_offset
	
	-- draw grid
	for y = 0, grid_size - 1 do		
		-- reset x_cell_count for each row
		x_cell_count = center_on_player and current.x - grid_offset or -grid_offset
		
		for x = 0, grid_size - 1 do
			
			local cleared = world:checkAlreadyCleared(x_cell_count, y_cell_count) -- world global
			
			local bossroom = false
			
			if math.abs(x_cell_count) - math.abs(y_cell_count) == 0 and x_cell_count ~= 0 and y_cell_count ~= 0 then 
					bossroom = true
			end
			
			
			if cleared then
				love.graphics.setColor(255, 207, 87, 255) -- cleared (gold)
			elseif bossroom then
				love.graphics.setColor(255,0,0,100) -- boss room (red)
			else
				love.graphics.setColor(255, 255, 255, 200) -- normal, not cleared (white)
			end
			
			if x_cell_count == current.x and y_cell_count == current.y then
				love.graphics.setColor(25, 255, 25, 255)  -- current zone (green)
			end
			
			love.graphics.rectangle('fill', left_start + x * (cell_size + cell_padding), top_start + y * (cell_size + cell_padding), cell_size, cell_size )
			love.graphics.setColor(255,255,255,255)
			
			
			x_cell_count = x_cell_count + 1
		end
		y_cell_count = y_cell_count + 1
	end
	
	
	
	
	
	love.graphics.setColor(255,255,255)
	-- GAME END
	if (player1.instance.ship.alive == false) then
		
		
		love.graphics.setFont(largeFont) -- SET FONT		
		if (game.newhiscore == true) then
			love.graphics.setColor(255,215,0)
			love.graphics.print('NEW HI SCORE', cw / 2 - 150, ch / 2 - 200)
		else
			love.graphics.setColor(178,34,34)
			love.graphics.print("YOU DIED", cw / 2 - 150, ch / 2 - 200)			
		end
		love.graphics.setColor(255,255,255)
		love.graphics.print("score : " .. player1.instance.score, cw / 2 - 150, ch / 2 - 140)
		love.graphics.print("press 'ESC' to try again", cw / 2 - 150,  ch / 2 - 80)
	end
	
	love.graphics.setFont(mainFont) -- SET FONT
	
	
	-- move FPS out of debug.
	love.graphics.print("FPS: " .. love.timer.getFPS(), 15, 15)
	
	--	DEBUG	
	
	if (game.settings.debug) then
	
		
		
		local pickup_count, nme_count = 0, 0
		for k,v in pairs(entities) do
			if (v.type == "enemy") then
				nme_count = nme_count + 1
			elseif (v.type == "pickup") then
				pickup_count = pickup_count + 1
			end
		end
		love.graphics.print("pickups : " .. pickup_count, 15, 35)	
		
		
		local count = 0
		for k,v in pairs(bullets) do
			count = count + 1
		end
		love.graphics.print("Bullets : " .. count, 15, 55)	
			
		local wall_count = 0
		for k,v in pairs(world.entities.walls) do
			wall_count = wall_count + 1
		end
		love.graphics.print("Walls : " .. wall_count, 15, 75)
		
		local explo_count = 0
		for k,v in pairs(explosions) do
			explo_count = explo_count + 1
		end
		love.graphics.print("Explosions : " .. explo_count, 15, 95)
		
		local sounds = love.audio.getSourceCount()
		love.graphics.print("Sounds : " .. sounds, 15, 115)
			
		love.graphics.print("Enemies :  " .. nme_count, 15, 135)
		
		love.graphics.print("Volume : " .. math.round(love.audio.getVolume(), 2), 15, 155)
		
		--love.graphics.print("Next enemy in :  " .. math.round(nme_spawn_timer), 15, 175)
			
	end
	
end