local class = require 'lib.middleclass'


Pickup = class('Pickup')

function Pickup:initialize(x, y, size) 

	self.name = "pickup"
	
	if (_size == nil) then
		self.h = 40
		self.w = self.h
	else
		self.h = size
		self.w = size
	end 
	
	self.x = x
	self.y = y
	self.angle = 0
	self.speedX = 0
	self.speedY = 0
	self.health = 1
	self.pointvalue = 10
	
	self.color_loop_counter = 0
	self.color = {
						r = 255,
						g = 207, 
						b = 87
					}
	
end

function Pickup:draw(sh)
		
	local color = self:getColor()
	love.graphics.setColor(color.r, color.g, color.b) 
	sh:rotate(self.angle)
	sh:draw('fill')
	
	
end

function Pickup:update(sh, dt)
	
	
	local rotatespeed = 5 -- atm it doesnt really matter where i put this. it just spins the shape.
	self.angle = rotatespeed * dt
	
	self.color_loop_counter = self.color_loop_counter + dt
	if self.color_loop_counter > 32 then
		self.color_loop_counter = 0
	end
	
	local loopvalue = math.abs(self.color_loop_counter)
	local frequency = 1
	
	self.color.r    = math.sin(frequency*loopvalue + 0) * 127 + 128;
	self.color.g = math.sin(frequency*loopvalue + 2) * 127 + 128;
	self.color.b   = math.sin(frequency*loopvalue + 4) * 127 + 128;

	self:checkStageClear(sh, dt)
	
end

-- Float towards the player
function Pickup:checkStageClear(sh, dt)
	
	local stageClear = game:checkStageClear()
	

	if stageClear then
		local px, py = player1:center()
		local sx, sy = sh:center()
		local speed = (1000/math.dist(sx, sy, px, py) )* 500
		
		
		local angle = math.angle(sx, sy, px, py)
		local x = speed * math.cos(angle) * dt
		local y = speed * math.sin(angle) * dt
		
		sh:move(x, y)
		
	end


end

function Pickup:setColor()  

	local color = self:getColor()
	love.graphics.setColor(color.r, color.g, color.b)

end

function Pickup:getColor()

	local color = {}
	color.r = self.color.r
	color.g = self.color.g
	color.b = self.color.b
	
	
	return color
end

 
 
  
 function Pickup:die(i, sh) -- index, shape
				
	-- apply buff to player
	-- select weapon for random bonus
	-- this is better now that i cleaned up modifier code
		
		local px, py = player1:center()
		local g = math.random (1,5)
		local w
		local buffs
		local announce_msg
		
		local r = math.random(1,5) -- random select buff
		
		local mod_props = WeaponModifiers:getPropNames()
		
		if g == 1 then 
			announce_msg = "Gun +"
			announce_color = "white"
			w = player1.instance.weaponguns
			w.modifiers:apply(w.weapon_type, mod_props[r])
			
		 elseif g == 2 then
			 announce_msg = "Bomb +"
			 announce_color = "blue"
			 w = player1.instance.weaponbomb
			w.modifiers:apply(w.weapon_type, mod_props[r])
			 
		 end
				
		if g <= 2 then
					
			--cobble together a decent message for player
			
			if r == 1 then
				announce_msg = announce_msg .. "Speed"
			elseif r == 2 then
				announce_msg = announce_msg .. "Range"
			elseif r == 3 then
				announce_msg = announce_msg .. "Damage"
			elseif r == 4 then
				announce_msg = announce_msg .. "Rate of Fire"
			elseif r == 5 then
				announce_msg = announce_msg .. "Bounce" 
			end
			
		elseif g == 3 then
			
			if (player1.instance.ship.repel.count < player1.instance.ship.repel.maxcount) then
				player1.instance.ship.repel.count = player1.instance.ship.repel.count + 1
				player1.instance.weaponrepel.armed = true
			end
			announce_msg = "Repel Aquired"
			announce_color = "gold"
			
		elseif g == 4 then
			
			local healthbonus = 10 -- * game:getStageModifier() -- NO SCALING ON THESE IT GETS RIDICULOUS
			if (player1.instance.health + healthbonus > player1.instance.maxhealth) then
				--player1.instance.maxhealth = player1.instance.health + healthbonus
				player1.instance.health = player1.instance.maxhealth
				
				announce_msg = "Health MAX"
				announce_color = "green"
			else
				player1.instance.health = player1.instance.health + healthbonus
				
				announce_msg = "+" .. tostring(healthbonus) .. " Health"
				announce_color = "green"
				
			end
			

		
		elseif g == 5 then
		
			if (player1.instance.ship.shield.timermax > 1.5) then
				player1.instance.ship.shield.timermax = player1.instance.ship.shield.timermax - 0.025
				announce_msg = "+Shield Cooldown"
				announce_color = "green"
			else
				announce_msg = "Shield Cooldown MAX"
				announce_color = "green"
			end
					
		end
		
		-- play sound 
		self:playSound()


		local font_w = mainFont:getWidth(announce_msg) -- not sure if this detects scale too. seems to return half as wide as i sould expect.
		game:spawnAnnouncement(px - font_w, py - 110, announce_color, announce_msg)
		
		
		-- apply point value to player
		player1.instance.score = player1.instance.score + self.pointvalue
		
		-- remove PICKUP entity
		Collider:remove(sh)
		table.remove(entities, i)
		
 end
 
 function Pickup:resolveCollision(sh, dx, dy)

end


function Pickup:playSound()
	-- should use a table to track this particular sound effect and limit to 5 or something. 
	-- as of now with a setting of 10 it is still irritating, but if i set it lower, sometimes it wont play at all due to bullet sfx driving up the sounds count.
	
	local sounds = love.audio.getSourceCount() -- on later stages you could have dozens flying back at the same time. no need to hammer them with this sound.
	if sounds <= 10 then -- have to account for bullets and explosions. buffed it up from 5 sounds.
			if sfx.wrench:isPlaying() then
				local newSource = sfx.wrench:clone()
				newSource:play()
			else
				sfx.wrench:play()	
		end
	end
	
end

   
return Pickup