local class = require 'lib.middleclass'

local custom_maps = {}
custom_maps[1] = require 'custom_maps.map1'



World = class('World')

function World:initialize(w, h, rows, columns)


	self.w = w
	self.h = h
	self.rows = rows
	self.columns = columns
	
	-- increase enough so that player reaching the border will not be uncentered from the camera.
	self.bufferpixels = 5000
	self.borderwidth = 100
	
	self.left = 0
	self.top = 0
	
	-- this variable name is misleading. this causes the leveltimer to be set, which calls toggleBorders() 
	self.bordersunlocked = false 
	
	self.showclearedmessage = false
	
	self.walltileopacitydefault = 100
	self.walltileopacity = 100
	
	self.currentstage = { x = 0, y = 0}
	self.stagescleared = {}
	
	self.entities = {}
	self.entities.walls = {}
	self.entities.stars = {}
	
	self.level_tiles = {}
	-- call init functions
	--self:generate() -- if i call this here, then levels.lua cant access world properties (world.w, world.h)
							-- moved to game.lua
	
end




function World:generate()
	local seed = math.round(self.left) + math.round(self.top)
	math.randomseed(seed + game.startseed) -- should keep all random world generation results the same for each play session
	
	-- reset wall tile opacity for level
	self.walltileopacity = self.walltileopacitydefault
	--reset bodersUnlocked status
	self.bordersunlocked = false
	self.showclearedmessage = false
	
	
	local boss_stage = false
	if player1 then
		local xq = math.abs(player1.instance.xquadrant) -- this is shit, need to track "zone"/quadrant better.
		local yq = math.abs(player1.instance.yquadrant) 
		if xq - yq == 0 and game:getStageModifier() > 1 then -- check stage modifier so that we dont get boss on reentering 0,0
			boss_stage = true
		end
		
		-- set current stage 
		self.currentstage.x = player1.instance.xquadrant
		self.currentstage.y = player1.instance.yquadrant
	end
	
	
	
	local alreadycleared = self:checkAlreadyCleared()
	
	
	
	if not alreadycleared then
	
		if not boss_stage then -- NORMAL STAGE
			-- define tile size
			self.rows = math.random(5,13)
			self.columns = self.rows + 7
			-- Inititalize levels
			local levels_src = Levels(self.rows, self.columns)
			self.level_tiles = levels_src:generate() -- generate stage tiles procedurally

		else -- ELSE CUSTOM LEVEL
			
			local custom = custom_maps[1]
			local cr = 0
			for i, v in ipairs(custom) do  -- i think this is unecessary for count
				cr = i
			end
			
			self.rows = cr 
			self.columns = #custom[1]

			local levels_src = Levels(self.rows,self.columns)
			self.level_tiles = levels_src:generateCustom(custom) -- custom level
		end
		
	else
	
		
		-- generate empty stage
		local levels_src = Levels(self.rows, self.columns)
		self.level_tiles = levels_src:generateEmpty()
		
	end
	
	
	self:buildStars()
	self:buildWalls(self.level_tiles, boss_stage)
	
	-- no entities if cleared
	if not alreadycleared then
		self:addEntities(boss_stage)
	end
	
end


function World:buildWalls(tiles, boss_stage) -- boss_stage:boolean

	--flag whether to remove walls at random on generation
	local wallRemoval = true
	if boss_stage then
		wallRemoval = false
	end

	-- RESET 
	for i, v in ipairs(self.entities.walls) do
		--print ('removing')
		Collider:remove(v)
	end
	self.entities.walls = {}

	-- BORDER WALLS / TELEPORTERS??
	local bw = self.borderwidth -- border width
	
	local left = Collider:addRectangle(-bw + self.left, -bw + self.top, bw, self.h + bw*2)
	Collider:addToGroup("group_walls",left)
	left.type = "wall"
	left.subtype = "border"
	left.collision_effect = nil
	left.warp_dir = "left"
		
	local top = Collider:addRectangle(0 + self.left, -bw  + self.top, self.w, bw)
	Collider:addToGroup("group_walls",top)
	top.type = "wall"	
	top.subtype = "border"
	top.collision_effect = nil
	top.warp_dir = "top"
		
	local right = Collider:addRectangle(self.w + self.left, -bw + self.top, bw, self.h + bw*2)
	Collider:addToGroup("group_walls",right)
	right.type = "wall"
	right.subtype = "border"
	right.collision_effect = nil
	right.warp_dir = "right"
		
	local bottom = Collider:addRectangle(0 + self.left, self.h + self.top, self.w, bw)
	Collider:addToGroup("group_walls",bottom)
	bottom.type = "wall"
	bottom.subtype = "border"
	bottom.collision_effect = nil
	bottom.warp_dir = "bottom"
		
	table.insert(self.entities.walls, left)
	table.insert(self.entities.walls, right)
	table.insert(self.entities.walls, top)
	table.insert(self.entities.walls, bottom)
	
	-- create interior walls
	local w = self.w / self.columns
	local h = self.h / self.rows

		--print("GENERATING : ", self.rows, self.columns)
	
	for y = 0, self.rows -1 do
		for x = 0, self.columns -1  do
		
			
			
			if tiles[y][x] == "#" then 
				if ( math.random(1,10) > 3 or wallRemoval == false) then
					local wall = Collider:addRectangle(x*w + self.left, y*h + self.top, w, h)
					wall.type = "wall"
					wall.count = count
					Collider:addToGroup("group_walls", wall)
					Collider:setPassive(wall) -- passive shapes won't actively search for collisions, but will collide with active shapes.
					--Collider:setGhost(wall)
					table.insert(self.entities.walls, wall)
				else				
					tiles[y][x] = "R" -- removed wall with random above. || is this assignment changing self.level_tiles??
				end
				
				
			end
			
			--io.write(tostring(tiles[y][x]))		
			
		end
		--io.write("\n")
	end
	
	--print('length of first row of map tiles ' .. #tiles[0])
	
end


function World:drawWalls()
	
	-- draw walls
	for i, v in ipairs(self.entities.walls) do
	local drawself = true
	
		if i < 5 then 
		
			if v.collision_effect == "warp" then
				love.graphics.setColor(255, 207, 87) -- borders will warp
			else
				love.graphics.setColor(79, 124, 176) -- borders are locked
			end
			
		else
			love.graphics.setColor(65, 49, 119, self.walltileopacity) -- interior walls
			
			-- don't draw interior walls if they are not in camera
			-- increase area of camera to account for pop-in of walls on the primary camera
			local tilesize = (self.w / self.rows) /2
			local l,t,w,h = game.cam1:getVisible()
			local x1, y1 = l - tilesize, t - tilesize
			local x2, y2 = l + w + tilesize, t + h + tilesize
			local wall_x, wall_y = v:center()
			
			if not point_contained(x1, y1, x2, y2, wall_x, wall_y) then			
				drawself = false
			end		
		end
		

		if (drawself) then
			v:draw('fill')
		end
	end
	
end

function World:buildStars()

	--clear any stars
	self.entities.stars = {}

	local totalstars = math.random(500,1000) 
	for i = 0, totalstars, 1 do
		table.insert( self.entities.stars, 
								{ 
									x = math.random(self.w) + self.left, 
									y = math.random(self.h) + self.top, 
									alpha = math.random(1,255)
								} 
						)
	end

end

function World:drawBackground() 
	
	--give the stage a black background
	love.graphics.setColor(0,0,0,100)
	love.graphics.rectangle('fill', self.left, self.top, self.w, self.h)
	
	--draw stars	
	
	love.graphics.setLineWidth(1)
	for i, v in ipairs(self.entities.stars) do
		love.graphics.setColor(255,255,255, v.alpha)
		love.graphics.point(v.x, v.y)
	end
	
		
		-- DRAW STAGE CLEARED TEXT ON STAGE
	if self.showclearedmessage then
	
		if self:checkAlreadyCleared() then
		local l, t, w, h = game.cam1:getWorld()
		local text_l = l + self.bufferpixels/2 + self.w / 2
		local text_t = t + self.bufferpixels/2 + self.h / 2
		
		local line1_text = "STAGE"
		local line2_text = "CLEARED"
		
		love.graphics.setFont(giganticFont)
				
		love.graphics.setColor(255,207,87,255)
		love.graphics.print(line1_text, text_l - 500, text_t - 600) -- i should use stage width/2 height/2 and then hardcoded numbers..
		love.graphics.print(line2_text, text_l - 700, text_t - 200 ) -- ..incase i make a very tall or wide stage by defining world size differently.
		love.graphics.setFont(mainFont)
		end
		
	end
	
  
end



function World:updateWorld(dt)

	local px, py = player1:center()
	local l,t,w,h = game.cam1:getWorld()

	local stageclear = game:checkStageClear()
	
	if stageclear then		
		self:stageCleared(dt)
	end
	
	local warped = false
	
	if (py < self.top) then
		game.cam1:setWorld(l, t - h , w, h)		
		self.top = self.top - h		
		warped = true
	end
	
	if (py > self.top + h) then
		game.cam1:setWorld(l, t+h, w, h)
		self.top = self.top + h
		warped = true
	end
	
	if (px < self.left) then
		game.cam1:setWorld(l - w, t, w, h)
		self.left = self.left - w
		warped = true
	end
	
	if (px > self.left + w) then
		game.cam1:setWorld(l + w, t, w, h)
		self.left = self.left + w
		warped = true
	end
	
	
	if warped then
		Entities:resetAll()
		self:generate()		
	end
	
end


function World:toggleBorders()


	for i, v in ipairs(self.entities.walls) do
		if i < 5 then 
			if v.collision_effect == "warp" then
				v.collision_effect = nil
			else
				v.collision_effect = "warp" 
				self.showclearedmessage = true
			end
		end
	end
	
	
	
end


function World:getTileXY(row, col) 
	
	col = col + 1
	row = row + 1

	local w = math.round(self.w / self.columns ) -- 1 column width
	local h = math.round(self.h / self.rows ) -- 1 row width
	
	-- x = col, y = row
	local x = ((col * w) - (w/2)) + world.left -- world.left discounts buferpixels. i could probably fix getTilePosition to not use player quadrants similarily
	local y = ((row * h) - (h/2)) + world.top -- same as above.

	--print ('returning x, y : ' .. x, y)
	
	return x, y
end


-- global==true means full grid. 
-- global == false means relative to initial world boundaries.. hopefully
function World:getTilePosition(_x, _y, global) 
	
	--print ("current x, y: ".. _x .. ", " .. _y)
	local w = math.round(self.w / self.columns ) -- 1 column width
	local h = math.round(self.h / self.rows ) -- 1 row width
	local x, y
	
	if (global) then -- adjust for grid position. 
		x = _x - ((player1.instance.xquadrant) * (self.w + self.bufferpixels)) -- adjust for global x
		y = _y - ((player1.instance.yquadrant) * (self.h + self.bufferpixels)) -- adjust for global y
	else
		x = _x
		y = _y
	end
	
	local col = math.ceil((x / w) ) - 1 
	local row = math.ceil((y / h) ) - 1	

	return row, col 
	
end

function World:getTileValue(row, col)

	--print (row, col)

	--row/col (y,x) instead of col/row (x,y) because map axis are reversed 
	local tile = self.level_tiles[row][col]

	return tile
end

function World:addEntities(boss_stage)
	local modifier = game:getStageModifier() 
		
	
	if not boss_stage then
		for i = 1, modifier do
				
				local spawned = false
				
				-- at least 1 must spawn
				while not spawned do
					-- 50% chance of spawn each loop
					if math.random(1,2) > 1 then
						game:addEntity('enemy1')
						spawned = true
					end
					
					-- 50% chance of spawn each loop
					if math.random(1,2) > 1 then
						game:addEntity('enemy2')
						spawned = true
					end

					-- 50% chance of spawn each loop
					if math.random(1,2) > 1 then -- enemy 3 cannot trigger spawned flag. we dont want only these guys to spawn.
						game:addEntity('enemy3')
						--spawned = true
					end

				end
		end
		
		
		local pickupspawns = math.floor(10/modifier) -- once you get a certain distance out, you'll have plenty of pickups from killing enemies.
		for i = 1, pickupspawns do
			game:addEntity('pickup')
			
		end
	else	
	
		-- must add boss here it is boss stage
		game:addEntity('boss1')
	end
	
	-- if stage modifier is over 10, start adding boss1 to stages (including boss stage)
	if modifier >= 10 then
			local bosses_to_add = math.floor(modifier/10) 
			--print('math.floor('..tostring(modifier)..'/5) == ', bosses_to_add)
			for i = 1, bosses_to_add do
							
				-- 50% chance of spawn each loop
				if math.random(1,2) > 1 then
					game:addEntity('boss1')
				end 
				
			end
	end

end


-- if stage is cleared, enable stage unlock timer, and slowly make walls disappear
function World:stageCleared(dt)

	if not self.bordersunlocked then	
		game.leveltimer = game.levelduration -- initiate short countdown before borders unlock
		self.bordersunlocked = true -- prevent this conditional from triggering more than once.
		
		-- add current stage to stages cleared
		if not self:checkAlreadyCleared() then
			local stagex = self.currentstage.x
			local stagey = self.currentstage.y
			table.insert(self.stagescleared, {x = stagex, y = stagey})
		end
	end

	if self.walltileopacity > 0 then 
		self.walltileopacity = self.walltileopacity - 1
	else
		
		-- remove walls from collider once they are fully invisible.
		for i, v in ipairs(self.entities.walls) do
			if i > 4 then  -- not borders
				Collider:setGhost(v) -- set wall to ghost until player leaves zone, then walls will be destroyed.
			end		
		end
		
		-- remove remaining enemy3's now that stage is cleared of attacking enemies.
		for _, v in pairs(entities) do
		if (v.name == "enemy3") then				
			v.instance.health = 0
		end		
	end
		
		
	end
	
end


-- checks if stage has already been cleared this playthrough.
function World:checkAlreadyCleared(col, row)

	local x = col or self.currentstage.x
	local y = row or self.currentstage.y

	local stagefound = false
	for _, v in ipairs(self.stagescleared) do	
		if v.x == x and v.y == y then
			stagefound = true
		end
	end
	
	return stagefound
end



return World